unit uoddface_calibraparque;

{rch@20150923
Definición del problema de Optimización de Modelo de Parque Eólico.
Se calibra el mdelo en función de series de medidas de

velocidad, dirección, humedad realtiva, temperatura, potencia MW.

Del parque se supone que se conocen las posiciones de los molinos y del anemo
estando el anemo en las coordenadas (0,0).

}


{$mode delphi}

interface

uses
  Classes, SysUtils,
  uoddface,
  xmatdefs,
  matreal,
  uConstantesSimSEE,
  uauxiliares,
  uDataSetGenerico,
  ucalibrador,
  urosx;

type

  TCalibraParque_Problema = class(TProblema)
    CarpetaSala: string;
    NombreSala: string;
    Calibrador: TCalibradorParque;

    constructor Create(dbconx: TDBrosxCon; recProblema: TDataRecord;
      idEjecutor_: integer; tmp_rundir_: string);
    class function CreateFromDB(dbconx: TDBrosxCon; nid_problema: integer;
      idEjecutor_: integer; tmp_rundir_: string): TCalibraParque_Problema;

    function evaluar_(Individuo: TIndividuo; SemillaAleatoria: integer): boolean;
      override;
    procedure Free; override;


    procedure FijarIndividuo(individuo: TIndividuo; var EsFactible: boolean);
    procedure DumpCalibradorToArchi( nid: integer; archi: string );

  end;




implementation




constructor TCalibraParque_Problema.Create(
  dbconx: TDBrosxCon;
  recProblema: TDataRecord;
  idEjecutor_: integer; tmp_rundir_: string);

var
  kPar: integer;
  k, kMolino: integer;
  nombrePar: string;
  NArmonicos: integer;
  cnt_par: integer;
begin
  // Lee la información general del problema y baja el archivo de sala
  // en nuestro caso la sala es un TCalibrador
  inherited Create( dbconx, recProblema, idEjecutor_, tmp_rundir_);

  // calculamos carpeta de la sala y nombre
  //  CarpetaSala := extractFilePath( ArchiSala );
  CarpetaSala := tmp_rundir_;

  NombreSala := extractFileName(ArchiSala);
  // acomodamos el directorio de la sala
  ArchiSala := tmp_rundir_ + barraDir + NombreSala;

  Calibrador:= TCalibradorParque.CreateLoadFromArchi( archisala );

  // Bien, ahora que tenemos la información del problema  creamos
  // los descriptores de ADN

  setlength(DescriptoresE, 0 );
//  setlength(DescriptoresR,  length( Calibrador.paramsPtr ) );
  setlength(DescriptoresR,  Calibrador.parque_MT.Meteo_etha.n + 1 );

  cnt_par:= 0;

  DefinirParametroReal( cnt_par, 'ME_A0', 0.5, 2, 8 );
  inc( cnt_par );

  for k:= 1 to ( Calibrador.parque_MT.Meteo_etha.n -1 ) div 2 do
  begin
    DefinirParametroReal( cnt_par, 'ME_xa'+IntToStr( k ), -1, 1, 8 );
    inc( cnt_par );
  end;
  for k:= 1 to (Calibrador.parque_MT.Meteo_etha.n -1 ) div 2 do
  begin
    DefinirParametroReal( cnt_par, 'ME_xb'+IntToStr( k ), -1, 1, 8 );
    inc( cnt_par );
  end;

  (*
  DefinirParametroReal( cnt_par, 'alfa_terreno', 0.001, 2, 8 );
  inc( cnt_par );
  DefinirParametroReal( cnt_par, 'agamma_atmosfera', 0.001, 2, 8 );
  inc( cnt_par );
  DefinirParametroReal( cnt_par, 'fpv3pe', 1, 3, 8 );
  inc( cnt_par );
  *)
  DefinirParametroReal( cnt_par, 'Factor_Perdidas', 0.001, 0.2, 8 );
  inc( cnt_par );



  (***
  for k := 0 to high( Calibrador.parque_MT.Molinos ) do
  begin
    DefinirParametroReal( cnt_par, 'Etha_Molino_'+IntToStr( k+1 ), 0.5, 2, 8 );
    inc( cnt_par );
  end;
      ***)
  CalcularLargosADN;
end;

class function TCalibraParque_Problema.CreateFromDB(
  dbconx: TDBrosxCon;
  nid_problema: integer;
  idEjecutor_: integer; tmp_rundir_: string): TCalibraParque_Problema;
var
  ds: TResultadoQuery;
  r: TDataRecord;
  ap: TCalibraParque_Problema;
begin
  ds := dbconx.sql_query('SELECT * FROM ofe_problemas WHERE nid = ' + IntToStr(
    nid_problema) + ' LIMIT 1 ');
  r := ds.Next;
  if r <> nil then
    ap := TCalibraParque_Problema.Create(dbconx, r, idEjecutor_, tmp_rundir_)
  else
    ap := nil;
  ds.Free;
  Result := ap;
end;


function nextpal_sep(var r: string; sep: string): string;
var
  res: string;
  i: integer;
begin
  r := trim(r);
  i := pos(sep, r);
  if i > 0 then
  begin
    res := trim(copy(r, 1, i - 1));
    Delete(r, 1, i);
  end
  else
  begin
    res := r;
    r := '';
  end;
  Result := res;
end;


function nextpal_tab(var r: string): string;
begin
  Result := nextpal_sep(r, #9);
end;


procedure readln_vect(var f: textfile; var v: TDAOfNreal; N: integer);
var
  r: string;
  pal: string;
  k: integer;
begin
  setlength(v, N);
  system.readln(f, r);
  pal := nextpal_tab(r); // caption
  for k := 0 to high(v) do
  begin
    pal := nextpal_tab(r);
    v[k] := StrToFloat(pal);
  end;
end;







procedure TCalibraParque_Problema.FijarIndividuo(individuo: TIndividuo; var EsFactible: boolean);
var
  kPar: integer;

  jArmonico, NArmonicos: integer;
  m, A0, resto, p: NReal;
  Ak, Bk: NREal;
  j: integer;
  cnt_par: integer;
begin

  cnt_par:= 0;
  m:= 0;
  NArmonicos:=  (Calibrador.parque_MT.Meteo_etha.n-1) div 2;
  A0 := individuo.XR.e( cnt_par+1 );
  Calibrador.params_xo[cnt_par]:=  A0;
  inc( cnt_par );
  resto:= 0.5 * A0;
  for jArmonico := 1 to NArmonicos do
  begin
    Ak:=  individuo.XR.e( cnt_par+1 )/sqrt(2);
    Bk:=  individuo.XR.e( cnt_par+1+NArmonicos )/sqrt(2);
    m:= sqrt( sqr( Ak )+ sqr( Bk ) );
    p:= m * resto;
    Calibrador.params_xo[cnt_par]:= Ak*p;
    Calibrador.params_xo[cnt_par+NArmonicos]:= Bk*p;
    inc( cnt_par );
    resto:= resto - p;
  end;
  inc( cnt_par, NArmonicos );

  for j:= 1 to 1 do
  begin
    Calibrador.params_xo[cnt_par]:= individuo.XR.e( cnt_par + 1 );
    inc( cnt_par );
  end;

  (*
  for j:= 2+NArmonicos*2+1  to individuo.XR.n do
    Calibrador.params_xo[j-1]:= individuo.XR.e(j);
    *)
  EsFactible:= true;
end;



procedure TCalibraParque_Problema.DumpCalibradorToArchi( nid: integer; archi: string );
var
  a: TIndividuo;
  EsFactible: boolean;
begin
  a := LeerIndividuo(nid);
  decodificar_adn(a);
  FijarIndividuo( a, EsFactible );
  Calibrador.SetParams( Calibrador.params_xo );
  Calibrador.StoreInArchi( archi );
end;


function TCalibraParque_Problema.evaluar_(Individuo: TIndividuo;
  SemillaAleatoria: integer): boolean;
var
  EsFactible: boolean;
  costo: NReal;
begin
  FijarIndividuo(Individuo, EsFactible);
  if EsFactible then
  begin
    costo := Calibrador.Costo( Calibrador.params_xo, SemillaAleatoria );
    Individuo.f_Objetivo := costo;
    Result := True;
  end
  else
    Result := False;
end;



procedure TCalibraParque_Problema.Free;
begin
  if Calibrador <> nil then
    Calibrador.Free;
  inherited Free;
end;

end.
