unit uparques_modeloempirico;

{$mode delphi}
{$DEFINE SOLO_GRAFICA_CURVAS_S}
interface

uses
  Classes, SysUtils, Math,
  {$IFDEF CON_GRAFICOS}
  Graphics, ugraficador,
  {$ENDIF}
  uauxiliares,
  MatEnt,
  uro_aire, umodelosintcegh, xmatdefs, matreal, ucaminos_modelado_ro,
  uparques_modeloteorico, uiterador_laplaciano_nulo, ucurvasaeros;

const
  UmbralPoblacionMinimaCluster = 10;
  NClusterSectores = 72;
  NClusterTramosVelocidad = 10;
  OverllapingClusterSectores = 0.2;

  filt_dist_pu = 0.1;


type

  { TParque_ModeloEmpirico }

  TParque_ModeloEmpirico = class
    Parque_MT: TParque_ModeloTeorico; // puntero a modelo Teórico.

    // ***************grupo Series de datos**************************
    // Este grupo de variables solo se cargan si el Parque se construye
    // desde Create( archiseries )
    NDatos: integer; // largo de la series de datos
    dtPrimerMuestra, dtEntreMuestras: TDateTime;
    series: TSeriesDeDatos; // las series históricas de datos
    // punteros a las series por nombre
    sDir_gr, sVel_mps, sPot_MW, sTemp_grC, sPre_hPa, sHumRel_p100, sDisp, sCGM,sI_Turb: TVectR;
    // Filtros de las series
    // van de 0..nDatos (el cero no se usa) un TRUE indica que el dato será filtrado.
    filtro_huecos_vdp, filtro_coherencia, filtro_cgm, filtro_distancia: TDAOfBoolean;

    // cantidad de huecos introducidos en la Potencia por Disp < 0.8
    cnt_filtrados_disp: integer;

    //************** fin del grupo Series de datos ******************

    // Factor de Memoria de la velocidad vista.
    Alfa_velFilt: NReal;

    // Cluster de muestras asociadas a un ángulo
    pv3Anemo_ClusterAng, pe_MW_ClusterAng,I_Turb_ClusterAng: TVectR;

    // matriz con una fila por angulo con las muestras Representantes.
    // los representantes son para pv3 = k * pv3(20 mps)/ 10   con k:= 0 .. 10
    // el peso_Representantes se pone el peso que se debe asignar a esa muestra
    // en la calibración del modelo. Un peso < 0 indica que esa muestra no debe
    // ser considerada
    pv3_Representados: TVectR;
    peso_Representantes, pe_MW_Representantes_a, pe_MW_Representantes_b,
    Representante_FO: TMatR;
    Representantes_pv3_medio: TVectR;
    curva_s_alfa, curva_s_vm, curva_s_errc2: TVectR;
    // valores por sector de la curva representativa del parque
    curva_s_cntMuestras: TVectE;

    constructor Create(parque: TParque_ModeloTeorico; archi_series: string);

    // Separación de las muestras.
    // Rellena los vectores: pv3Anemo_ClusterAng, pe_MW_ClusterAng
    // Con las muestras que pasan los filtros de huecos, coherencia y cgm
    // El vector pe_MW_ClusterAng, contiene los valores corregidos por la disponibilidad.
    procedure Cluster_Ang(ang0, dang: NReal);

    // Recorre las muestras del cluster y crea el grupo de representantes
    // para ser usados en el cálculo de la función objetivo
    procedure Representar_Clusters(NSectores: integer;
      NTramosRepresentados: integer; overlapping: NReal; cnt_Pasada: integer);

    function Calc_PE_hermit(u, dx: NReal; var Peso: NReal; kTramo: integer;
      rep_Peso, rep_F0, rep_b: TVectR): NReal;

    // Retorna la potencia eléctrica en MW en función del ángulo
    // de la dirección del viento en grados y de la potencia de la velocidad
    // del viento den el anemómetro. En la variable Peso retorna la cantidad
    // de muestras asociadas a la respuesta.
    function Calc_PE(ang, pv3Anemo, ro: NReal; var Peso: NReal): NReal;
    //Cálculo de PE utilizando las curvas_S
    function Calc_PE_curva_s(ang, pv3Anemo, ro: NReal): NReal;

    constructor Create_LoadFromFile(var f: textfile; parque: TParque_ModeloTeorico);
    procedure StoreInFile(var f: textfile);

    procedure Free;


    // Retorna la cantidad de Sectores representados.
    function NSectores: integer;

    // REtorna la cantidad de Tramos de Velocidad representados.
    // La cantidad de puntos es NTramosVel+1
    function NTramosVel: integer;

    // Retorna el ángulo en grados siendo jSector el ordinal del sector de 1..NSectores
    function AngOfjSector(jSector: integer): NReal;

    // Retorna el número de sector dado el ángulo en grados.
    // en kAng1 y kAng2 retorna los índices (base 1) de los representantes anterior y posterior a Ang
    // en peso1 y peso2 los pesos para ponderar los respectivos representantes en la posición Ang.
    procedure SectorOfAng(var kAng1, kAng2: integer; var peso1, peso2: NReal;
      Ang: NReal);

    // Escribe las curvas S
    procedure WriteCurvas_S_xlt;
  private

    // Hace chequeos mínimos de coherencia de las series y
    // filtra errores obvios. Es llamado en el create desde series
    // También se filtra los datos en los que actuó la CGM.
    procedure FiltradoDeCoherenciaYCGM;

    procedure Calc_Curva_s_cluser_ang(
      var curva_s_alfa, curva_s_vm, curva_s_errc2: NReal;
      var curva_s_cntMuestras: integer; pv3Anemo_ClusterAng, pe_MWClusterAng: TVectR;
      Parque_MT_Curva_s: TCurvaAeroGen);

    {$IFDEF CON_GRAFICOS}
    // escribe archivo seriesConFiltro.xlt
    // y gráficos de Muestras y Filtros
    procedure WriteSeriesConFiltros;

    // imprime graficas del cluster jSector
    procedure PLOT_GRAFICAS_DEL_CLUSTER_ANG(jSector: integer;
      pv3_Representados, pv3_medio: TVectR; muestras_pv3, muestras_PE, muestras_Turb: TDAofNReal;
      v_Pesos, v_a, v_b, v_F0: TVectR; alfa, vm: NReal; cnt_pasda: integer);
    {$ENDIF}
  end;

function detectar_fijos(var vector: TVectR;
  NRep: integer; ValoresExcluidos: TStringList): integer;

function detectar_fuera_de_rango(var vector: TVectR; min, max: NReal;
  ValoresExcluidos: TStringList): integer;

implementation


// Si InvertirFiltros = false, retorna un DAOfNReal con los valores que pasan todos los filtros.
// Si InvertirFiltros = true, retorna lso valores filtrados
// el Filtro_huecos indica hueco en los datos. Si Filtro_Huecos<>NIL NO se consideran
// los casilleros en que filtro_huecos[k] = TRUE (o sea, en la salida no se aparecen)
function AplicarFiltros(filtros: array of TDAOfboolean; v: TVectR;
  InvertirFiltros: boolean; filtro_huecos: TDAOfBoolean
  // poner a NIL si no se quiere aplicar
  ): TDAOfNReal;
var
  k: integer;
  cnt: integer;
  res: TDAOfNReal;
  flg_filtrar: boolean;
  kFiltro: integer;
  quitar_huecos: boolean;
begin
  setlength(res, v.n);
  cnt := 0;
  quitar_huecos := filtro_huecos <> nil;
  for k := 1 to v.n do
  begin
    if quitar_huecos and filtro_huecos[k] then
      continue;
    flg_filtrar := InvertirFiltros;
    for kFiltro := 0 to high(filtros) do
    begin
      if filtros[kFiltro][k] then
        flg_filtrar := not InvertirFiltros;
      break;
    end;
    if not flg_filtrar then
    begin
      res[cnt] := v.e(k);
      Inc(cnt);
    end;
  end;

  setlength(res, cnt);
  Result := res;
end;



procedure TParque_ModeloEmpirico.WriteCurvas_S_xlt;
var
  f: textfile;
  kAng: integer;
  ang: NReal;
begin

  assignfile(f, CarpetaResultados_modelado_RO(parque_MT.nid_Central) + '\curvas_s_.xlt');
  rewrite(f);
  writeln(f, 'ang', #9, 'alfa', #9, 'vm', #9, 'errc2', #9, 'cntMuestras',
    #9, 'MT_alfa', #9, 'MT_vm');

  for kAng := 1 to NSectores do
  begin
    ang := AngOfjSector(kAng);
    writeln(f, ang, #9, curva_s_alfa.e(kAng),
      #9, curva_s_vm.e(kAng), #9, curva_s_errc2.e(kAng),
      #9, curva_s_cntMuestras.e(kAng), #9, Parque_MT.Curva_s.curva_s_alfa,
      #9, Parque_MT.Curva_s.curva_s_vm);
  end;
  closefile(f);

end;

{$IFDEF CON_GRAFICOS}
procedure TParque_ModeloEmpirico.WriteSeriesConFiltros;
var
  f: textfile;
  k: integer;
  dt: TDateTime;
  sx, sy: TSerieGrafico;
  g: TGrafico;
  v: TDAOfNReal;
begin

  assignfile(f, CarpetaResultados_modelado_RO(
    parque_MT.nid_Central) + '\series_filtradas.xlt');
  rewrite(f);
  writeln(f, 'dt', #9, 'Vel[m/s]', #9, 'Ang[°]', #9, 'PE[MW]', #9,
    'Temp[°C]', #9, 'Pre[hPA]',
    #9, 'HumRel[%]', #9, 'CGM[MW]', #9, 'fDisp[p.u.]',
    #9, 'Filt_Huecos', #9, 'Filt_Coherencia', #9, 'Filt_CGM', #9, 'Filtro_Distancia');
  for k := 1 to NDatos do
  begin
    dt := dtPrimerMuestra + (k - 1) * dtEntreMuestras;
    Write(f, DateTimeToStr(dt));
    Write(f, #9, sVel_mps.e(k));
    Write(f, #9, sDir_gr.e(k));
    Write(f, #9, sPot_MW.e(k));
    Write(f, #9, sTemp_grC.e(k));
    Write(f, #9, sPre_hPa.e(k));
    Write(f, #9, sHumRel_p100.e(k));
    Write(f, #9, sCGM.e(k));
    Write(f, #9, sDisp.e(k));
    Write(f, #9, boolToInt(filtro_huecos_vdp[k]));
    Write(f, #9, boolToInt(filtro_coherencia[k]));
    Write(f, #9, boolToInt(filtro_cgm[k]));
    Write(f, #9, boolToInt(filtro_distancia[k]));
    writeln(f);
  end;
  closefile(f);



  // Grafico P_Vel (muestras y filtrados )
  g := TGrafico.Create('gfiltrados_pv', TG_DispersionXY, 0);
  g.Titulo := 'Muestras y filtrados (plano PE-Velocidad)';
  g.eje_x.titulo := 'Velocidad de viento [m/s]';
  g.eje_x.Forzar_vmin(0);
  g.eje_x.Forzar_vmax(25);

  g.eje_y.Forzar_vmin(0);
  g.eje_y.Forzar_vmax(10 * (trunc(self.Parque_MT.Potencia_Autorizada_MW / 10) + 1));

  g.eje_y.titulo := 'Potencia eléctrica [MW]';
  sx := TSerieGrafico.Create('vel', sVel_mps.toTDAOfNReal, nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('PE', sPot_MW.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clGreen;
  sy.RadioMarcador := 2;
  sy.AnchoLineaMarcador := 1;
  g.AddSerie(sy);

  // ahora ploteamos los filtrados por coherencia
  v := aplicarFiltros([filtro_coherencia], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_coherencia', v, nil,
    nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_coherencia], sPot_MW, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('PE_filtrados_por_Coherencia', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clFuchsia;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);

  // ahora ploteamos los filtrados por cgm
  v := aplicarFiltros([filtro_cgm], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_cgm', v, nil, nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_cgm], sPot_MW, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('PE_filtrados_por_cgm', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);


  //LD y JFP: ahora ploteamos los filtrados por distancia
  v := aplicarFiltros([filtro_distancia], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_dist', v, nil, nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_distancia], sPot_MW, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('PE_filtrados_por_dist', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clBlack;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);

  g.SaveJPG(CarpetaResultados_modelado_RO(Parque_MT.nid_Central) +
    '\muestras_filtrados_plano_pe_vel_'+IntToStr(trunc(Alfa_velFilt*100))+'.jpg', 2100, 1000);
  g.Free;



  // Grafico dir_vel (muestras y filtrados )
  g := TGrafico.Create('gfiltrados_dirvel', TG_DispersionXY, 0);
  g.Titulo := 'Muestras y filtrados (plano Dir-Velocidad)';
  g.eje_x.titulo := 'Velocidad de viento [m/s]';
  g.eje_x.Forzar_vmin(0);
  g.eje_x.Forzar_vmax(25);
  g.eje_y.titulo := 'Dirección [°]';
  g.eje_y.Forzar_vmin(0);
  g.eje_y.Forzar_vmax(360);
  sx := TSerieGrafico.Create('vel', sVel_mps.toTDAOfNReal, nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('Dir', sDir_gr.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clGreen;
  sy.RadioMarcador := 2;
  sy.AnchoLineaMarcador := 1;
  g.AddSerie(sy);

  // ahora ploteamos los filtrados por coherencia
  v := aplicarFiltros([filtro_coherencia], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_coherencia_', v, nil,
    nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_coherencia], sDir_gr, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('Dir_filtrados_por_Coherencia', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clFuchsia;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);

  // ahora ploteamos los filtrados por cgm
  v := aplicarFiltros([filtro_cgm], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_cgm', v, nil, nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_cgm], sDir_gr, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('Dir_filtrados_por_cgm', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);

  //LD y JFP: ahora ploteamos los filtrados distancia
  v := aplicarFiltros([filtro_distancia], sVel_mps, True, filtro_huecos_vdp);
  sx := TSerieGrafico.Create('vel_filtrados_por_distancia', v, nil, nil, '##.##', True);
  g.AddSerie(sx);

  v := aplicarFiltros([filtro_distancia], sDir_gr, True, filtro_huecos_vdp);
  sy := TSerieGrafico.Create('Dir_filtrados_por_distancia', v, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clBlack;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);

  g.SaveJPG(CarpetaResultados_modelado_RO( Parque_MT.nid_Central) +
    '\muestras_filtrados_plano_dir_vel_'+IntToStr(trunc(Alfa_velFilt*100))+'.jpg', 2100, 1000);
  g.Free;

end;

{$ENDIF}
function detectar_fijos(var vector: TVectR; NRep: integer;
  ValoresExcluidos: TStringList): integer;
var
  k1, k: integer;
  cnt_total, largoVect: integer;
  cnt: integer;
  vant: NReal;
  buscando: boolean;
begin
  k1 := 1;
  cnt_total := 0;
  cnt := 0;

  largoVect := vector.n;

  buscando := True;
  while buscando and (k1 < largoVect) do
    if ((ValoresExcluidos.IndexOf(IntToStr(trunc(vector.e(k1)))) <> -1)) then
      Inc(k1)
    else
      buscando := False;


  if buscando then
  begin
    Result := 0;
    exit;
  end;

  vant := vector.e(k1);
  k := k1 + 1;
  while k < vector.n do
  begin
    if (vector.e(k) = vant) then
    begin
      Inc(cnt);
      if cnt >= nRep then
        while k1 < k do
        begin
          vector.pon_e(k1, -888888);
          Inc(k1);
        end;
    end
    else
    begin
      vant := vector.e(k);
      k1 := k;
      if cnt >= nRep then
        cnt_total := cnt_total + cnt;
      cnt := 0;
    end;
    Inc(k);
  end;
  Result := cnt_total;
end;

function detectar_fuera_de_rango(var vector: TVectR; min, max: NReal;
  ValoresExcluidos: TStringList): integer;
var
  k: integer;
  cant_fuera_rango: integer;
begin

  cant_fuera_rango := 0;
  for k := 1 to vector.n do
    if (abs(vector.e(k)) >= maxLongint) or
      ((ValoresExcluidos.IndexOf(IntToStr(trunc(vector.e(k)))) = -1) and
      ((vector.e(k) < min) or (vector.e(k) > max))) then
    begin
      Inc(cant_fuera_rango);
      vector.pon_e(k, -66666);
    end;

  Result := cant_fuera_rango;
end;


function detectar_iguales(var filtrar: TDAOfBoolean; v: TVectR;
  nRep: integer; const filtro_huecos: TDAOfBoolean
  // poner NIL si no se quiere considerar
  ): integer;
var
  k1, k: integer;
  cnt_total: integer;
  cnt: integer;
  vant: NReal;
  buscando: boolean;
begin
  k1 := 1;
  cnt_total := 0;
  cnt := 0;
  if filtro_huecos <> nil then
  begin
    buscando := True;
    while buscando and (k1 < high(filtro_huecos)) do
      if filtro_huecos[k1] then
        Inc(k1)
      else
        buscando := False;
  end;

  if buscando then
  begin
    Result := 0;
    exit;
  end;

  vant := v.e(k1);
  k := k1 + 1;
  while k < v.n do
  begin
    if ((filtro_huecos = nil) or (not filtro_huecos[k])) and (v.e(k) = vant) then
    begin
      Inc(cnt);
      if cnt >= nRep then
        while k1 < k do
        begin
          filtrar[k1] := True;
          Inc(k1);
        end;
    end
    else
    begin
      vant := v.e(k);
      k1 := k;
      if cnt >= nRep then
        cnt_total := cnt_total + cnt;
      cnt := 0;
    end;
    Inc(k);
  end;
  Result := cnt_total;
end;


procedure TParque_ModeloEmpirico.FiltradoDeCoherenciaYCGM;
var
  max_cgm, cgm99, cgm: NReal;
  max_PE: NReal;
  max_fd: NReal;

  k: integer;
  cnt: integer;

begin
  max_cgm := Parque_MT.Potencia_Autorizada_MW;
  cgm99 := max_cgm * 0.99;

  max_PE := sPot_MW.maxVal;

  if (max_PE > max_cgm * 1.20) then
    raise Exception.Create('max_cgm: ' + FloatToStr(max_cgm) +
      ' * 1.2 < Max_PE: ' + FloatToStr(max_PE));

  max_fd := sDisp.maxVal;
  if max_fd > 1 then
  begin
    writeln('WARNING!!!: max_fd: ' + FloatToStr(max_fd) +
      ' Asumo que es cantidad de máquinas y divido por este valor.');
    sDisp.PorReal(1.0/max_fd);
  end;


  cnt_filtrados_disp:=0;
  for k:= 1 to sPot_MW.n do
  if sDisp.e(k) <0.8 then
  begin
    sPot_MW.pon_e(k, -115000 );
    Inc(cnt_filtrados_disp);
  end
  else
    sPot_MW.pon_e(k, sPot_MW.e(k)/sDisp.e(k) );



  // los defino de 1 más para poder indexar de 1 a n como en los TVectR que filtran
  // el casillero [0] NO se utiliza.
  setlength(filtro_coherencia, sVel_mps.n + 1);
  setlength(filtro_cgm, sVel_mps.n + 1);
  setlength(filtro_huecos_vdp, sVel_mps.n + 1);
  setlength(filtro_distancia, sVel_mps.n + 1);

  vclear(filtro_coherencia);
  vclear(filtro_cgm);
  vclear(filtro_huecos_vdp);
  vclear(filtro_distancia);

  cnt := 0;
  for k := 1 to high(filtro_huecos_vdp) do
  begin
    if (sVel_mps.e(k) < -1000) or (sDir_gr.e(k) < -1000) or
      (sPot_MW.e(k) < -1000) or (sPre_hPa.e(k) < -1000)
      or (sHumRel_p100.e(k) < -1000)or (sTemp_grC.e(k) < -1000) then
    begin
      filtro_huecos_vdp[k] := True;
      Inc(cnt);
    end;
  end;
  writeln('Filtrado de huecos: ' + IntToStr(cnt));

  // ahora buscamos identificar tramos en que las variables están FIJAS
  // esos tramos corresponden a intervalos de falta de dato del SCADA
  writeln('Filtro vel: ', detectar_iguales(filtro_coherencia, sVel_mps,
    8, filtro_huecos_vdp));
  writeln('Filtro dir: ', detectar_iguales(filtro_coherencia, sDir_gr,
    8, filtro_huecos_vdp));
  writeln('Filtro Pot: ', detectar_iguales(filtro_coherencia, sPot_MW,
    8, filtro_huecos_vdp));

  // detectar actividad de CGM
  // filtramos las muestras  si la CGM < 0.99 * max_cgm sin importar la PE
  // porque el parque puede estar operando en un modo NO-Normal.
  cnt := 0;
  for k := 1 to sPot_MW.n do
    if (not filtro_huecos_vdp[k]) and (sCGM.e(k) < cgm99) then
    begin
      filtro_cgm[k] := True;
      Inc(cnt);
    end;

  if cnt > 0 then
    writeln('Se filtrarán ' + IntToStr(cnt) + ' por CGM  ');

  writeln('Filtrado por Disp < 0.8: ', cnt_filtrados_disp );

  //LD y JFP: WriteSeriesConFiltros;
end;

constructor TParque_ModeloEmpirico.Create(parque: TParque_ModeloTeorico;
  archi_series: string);
var
  dist_ME_Datos, filtro_distancia_aux: TVectR;
  PE_Dato, PE_ME, dir_Dato, pv3Anemo_Dato, ro_aire_Dato, vel_Dato: NReal;
  Temp_grC_Dato, Pre_hPa_Dato, HumRel_p100_Dato, Peso_Dato_: NReal;
  idx_Datos: TDAofNInt;
  cnt_filtrados, i, nroDato: integer;
  velFilt: NReal;
  k: integer;

begin
  inherited Create;
  Parque_MT := parque;
  series := TSeriesDeDatos.CreateFromArchi(archi_series);

  NDatos := series.series[0].n;
  dtPrimerMuestra := series.dtPrimeraMuestra;
  dtEntreMuestras := series.dtEntreMuestras;

  if (abs(dtEntreMuestras * (24 * 60) - 10.0) * 60) > 1 then
  begin
    raise Exception.Create('ERROR en archivo series, Perioedo Entre Muestras : ' +
      FloatToStr(series.PeriodoDeMuestreo_horas) + ' debe ser ' + FloatToStr(1.0 / 6.0));
  end
  else
    dtEntreMuestras := 10.0 / (24 * 60); // imponemos 10 minutos

  sVel_mps := series.series[0];


  alfa_velFilt:= 0; //0.4;

  for k:= 1 to sVel_mps.n do
    if sVel_mps.e(k) >-1 then
    begin
      velFilt:= sVel_mps.e(k);
      break;
    end;

  for k:= 1 to sVel_mps.n do
  begin
    if sVel_mps.e(k) >-1 then
    begin
      velFilt:= alfa_velFilt* velFilt + (1-alfa_velFilt) *sVel_mps.e(k);
      sVel_mps.pon_e(k, velFilt );
    end;
  end;


  sDir_gr := series.series[1];
  sPot_MW := series.series[2];
  sTemp_grC := series.series[3];
  sPre_hPa := series.series[4];
  sHumRel_p100 := series.series[5];
  sCGM := series.series[6];
  sDisp := series.series[7];
  sI_Turb := series.series[8];

  pv3Anemo_ClusterAng := nil;
  pe_MW_ClusterAng := nil;
  I_Turb_ClusterAng := nil;

  // Como efecto colateral corrige la serie de potencia con la disponibilidad
  FiltradoDeCoherenciaYCGM;
  NDatos := series.series[0].n;

  //LD y JFP: Acá se obtiene el primer ME
  Representar_Clusters(NClusterSectores, NClusterTramosVelocidad,
    OverllapingClusterSectores, 1);


  dist_ME_Datos := TVectR.Create_Init(NDatos);
  //LD y JFP:vector con distancias entre ME y datos del primer ME
  filtro_distancia_aux := TVectR.Create_Init(NDatos);
  filtro_distancia_aux.Ceros;

  for i := 1 to NDatos do
    dist_ME_Datos.pon_e(i, 99999);
  //LD y JFP: inicializo vector con 99999 para saber si corresponde a distancia real o filtro coherencia,cgm,etc.

  cnt_filtrados := 0;//guardo cantidad de datos filtrados por coherencia,cgm,etc.

  //LD y JFP: Se filtra el 10 % de los datos con mayor distancia al 1er ME obtenido.

  for nroDato := 1 to NDatos do
  begin
    if filtro_huecos_vdp[nroDato] or filtro_coherencia[nroDato] or
      filtro_cgm[nroDato] then
      //(Peso_Dato<10)  ????Se filtran tambien segun peso para no sacar datos que no se van a utilizar para calibrar?
    begin
      cnt_filtrados := cnt_filtrados + 1;
      continue;
    end;

    dir_Dato := self.sDir_gr.e(nroDato);
      vel_Dato := self.sVel_mps.e(nroDato);
    PE_Dato := self.sPot_MW.e(nroDato);


    Temp_grC_Dato := self.sTemp_grC.e(nroDato);
    Pre_hPa_Dato := self.sPre_hPa.e(nroDato);
    HumRel_p100_Dato := self.sHumRel_p100.e(nroDato);
    ro_aire_Dato := ro_aire(Temp_grC_Dato, Pre_hPa_Dato, HumRel_p100_Dato);

    pv3Anemo_Dato := pv3Ofv(vel_Dato, ro_aire_Dato);

    {$IFDEF USAR_HERMITE}
    PE_ME := self.Calc_PE(dir_Dato, pv3Anemo_Dato, ro_aire_Dato, Peso_Dato);

    if Peso_Dato > UmbralPoblacionMinimaCluster then
      dist_ME_Datos.pon_e(nroDato, abs(PE_ME - PE_Dato))
    else
      dist_ME_Datos.pon_e(nroDato, 0);
    {$ELSE}
    if (pv3Anemo_Dato>1e-3) and (PE_Dato<1)
      then WriteLn('Hola');
    PE_ME := self.Calc_PE_curva_s(dir_Dato,pv3Anemo_Dato,ro_aire_Dato);
    dist_ME_Datos.pon_e(nroDato, sqr(PE_ME - PE_Dato));
    {$ENDIF}
  end;


  SetLength(idx_Datos, NDatos + 1);
  for i := 1 to NDatos do
    idx_Datos[i] := i;

  dist_ME_Datos.Sort_idx(True, idx_Datos);  //ordeno segun distancia  para depurar


  for i := NDatos - cnt_filtrados downto trunc((NDatos - cnt_filtrados) * (1 - filt_dist_pu)) + 1 do
    //se filtra el 10 % del neto de datos totales-filtrados por otras causas
    filtro_distancia_aux.pon_e(i, 1);


  filtro_distancia_aux.UnSortByIdx(idx_Datos); //restauro orden modificado para depurar


  for i := 1 to NDatos do
    if abs(filtro_distancia_aux.e(i) - 1) < 1e-8 then
      filtro_distancia[i] := True;


  //LD y JFP: Acá se obtiene el ME definitivo luego de depuradas las muestras por distancia
  Representar_Clusters(NClusterSectores, NClusterTramosVelocidad,
    OverllapingClusterSectores, 2);



  {$IFDEF CON_GRAFICOS}
  WriteSeriesConFiltros;
  {$ENDIF}
  WriteCurvas_S_xlt;
  WriteLn('Fin err2CurvasS: ', curva_s_errc2.normEuclid, ' AlfaFilt: ', Alfa_velFilt );

end;

constructor TParque_ModeloEmpirico.Create_LoadFromFile(var f: textfile; parque: TParque_ModeloTeorico );
begin
  inherited Create;
  Parque_MT:= parque;
  setSeparadoresGlobales;

  setlength(filtro_cgm, 0);
  setlength(filtro_coherencia, 0);

  pv3Anemo_ClusterAng := nil;
  pe_MW_ClusterAng := nil;
  I_Turb_ClusterAng := nil;

  pv3_Representados := TVectR.CreateLoadFromFile(f);
  peso_Representantes := TMatR.CreateLoadFromFile(f);

  pe_MW_Representantes_a := TMatR.CreateLoadFromFile(f);
  pe_MW_Representantes_b := TMatR.CreateLoadFromFile(f);

  Representantes_pv3_medio := TVectR.CreateLoadFromFile(f);
  Representante_FO := TMatR.CreateLoadFromFile(f);


  curva_s_alfa:=  TVectR.CreateLoadFromFile(f);
  curva_s_vm:=  TVectR.CreateLoadFromFile(f);
  curva_s_cntMuestras:= nil;
  curva_s_errc2:= nil;

  setSeparadoresLocales;
end;

procedure TParque_ModeloEmpirico.StoreInFile(var f: textfile);
begin
  setSeparadoresGlobales;
  pv3_Representados.StoreInFile(f);
  peso_Representantes.StoreInFile(f);
  pe_MW_Representantes_a.StoreInFile(f);
  pe_MW_Representantes_b.StoreInFile(f);
  Representantes_pv3_medio.StoreInFile(f);
  Representante_FO.StoreInFile(f);
  curva_s_alfa.StoreInFile(f);
  curva_s_vm.StoreInFile(f);

  setSeparadoresLocales;
end;

procedure TParque_ModeloEmpirico.Free;
begin
  if series <> nil then
    series.Free;
  setlength(filtro_cgm, 0);
  setlength(filtro_coherencia, 0);

  if pv3Anemo_ClusterAng <> nil then
    pv3Anemo_ClusterAng.Free;
  if pe_MW_ClusterAng <> nil then
    pe_MW_ClusterAng.Free;
  if I_Turb_ClusterAng <> nil then
    I_Turb_ClusterAng.Free;
  if pv3_Representados <> nil then
    pv3_Representados.Free;
  if peso_Representantes <> nil then
    peso_Representantes.Free;
  if pe_MW_Representantes_a <> nil then
    pe_MW_Representantes_a.Free;
  if pe_MW_Representantes_b <> nil then
    pe_MW_Representantes_b.Free;
  if Representante_FO <> nil then
    Representante_FO.Free;
  if Representantes_pv3_medio <> nil then
    Representantes_pv3_medio.Free;

  if curva_s_alfa <> nil then
    curva_s_alfa.Free;
  if curva_s_vm <> nil then
    curva_s_vm.Free;
  if curva_s_errc2 <> nil then
    curva_s_errc2.Free;
  if curva_s_cntMuestras <> nil then
    curva_s_cntMuestras.Free;

end;


function TParque_ModeloEmpirico.NSectores: integer;
begin
  Result := pe_MW_Representantes_a.nf;
end;

function TParque_ModeloEmpirico.NTramosVel: integer;
begin
  Result := pv3_Representados.n - 1;
end;

function TParque_ModeloEmpirico.AngOfjSector(jSector: integer): NReal;
begin
  Result := (jSector - 1) / pe_MW_Representantes_a.nf * 360.0;
end;

procedure TParque_ModeloEmpirico.SectorOfAng(var kAng1, kAng2: integer;
  var peso1, peso2: NReal; Ang: NReal);
var
  rAng: NReal;
begin

  rAng := ang * NSectores / 360;
  kang1 := (trunc(rAng) mod NSectores) + 1;
  peso2 := frac(rAng);
  peso1 := 1 - peso2;
  kang2 := (kang1 mod NSectores) + 1;

end;



{$IFDEF CON_GRAFICOS}
procedure TParque_ModeloEmpirico.PLOT_GRAFICAS_DEL_CLUSTER_ANG(
  jSector: integer; pv3_Representados, pv3_medio: TVectR; muestras_pv3,
  muestras_PE, muestras_Turb: TDAofNReal; v_Pesos, v_a, v_b, v_F0: TVectR;
  alfa, vm: NReal; cnt_pasda: integer);
var
  sx, sy: TSerieGrafico;
  g: TGrafico;
  v, pe, pe_s,pe_s_Rep: TDAOfNReal;
  ang: NReal;
  NTramos: integer;
  k, j, idx_0a10, idx_10a20, idx_resto: integer;
  pv3, u, dx, peso, x: NReal;
  linea:TCajaGraficoXY;
  muestras_Turb_0a10,muestras_Turb_10a20,muestras_Turb_resto:TVectR;
  muestras_PE_0a10,muestras_PE_10a20,muestras_PE_Resto:TVectR;
  muestras_Pv3_0a10,muestras_Pv3_10a20,muestras_Pv3_Resto:TVectR;


begin

  Ang := AngOfjSector(jSector);

  // Grafico P_Vel (cluster )
  g := TGrafico.Create('gfiltrados_pv', TG_DispersionXY, 0);
  g.Titulo := 'Muestras cluster ang: ' + FloatToStrF(ang, fFFixed, 12, 1) + '°';
  g.eje_x.titulo := 'Potencia del viento [MW/m2]';
  g.eje_x.FloatFormat:= ffExponent;
  g.eje_x.FloatFormat_NDecimales:= 8;
  g.eje_x.FloatFormat_NDigitos:= 6;


  g.eje_y.titulo := 'Potencia eléctrica [MW]';
  g.eje_y.Forzar_vmax(  trunc( max( self.Parque_MT.Potencia_Autorizada_MW, self.Parque_MT.Potencia_Nominal )/10+1)*10 );
  g.eje_y.Forzar_vmin( 0 );

  //inicializar vectores!!!

  muestras_PE_0a10:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Pv3_0a10:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Turb_0a10:=TVectR.Create_Init(Length(muestras_Turb));


  muestras_PE_10a20:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Pv3_10a20:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Turb_10a20:=TVectR.Create_Init(Length(muestras_Turb));

  muestras_PE_Resto:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Pv3_Resto:=TVectR.Create_Init(Length(muestras_Turb));
  muestras_Turb_resto:=TVectR.Create_Init(Length(muestras_Turb));

  idx_0a10:=1;
  idx_10a20:=1;
  idx_resto:=1;

  //JFP: Separo los datos según su intensidad de turbulencia
  for k:=0 to Length(muestras_Turb)-1 do
    if (muestras_Turb[k]>0) and (muestras_Turb[k]<0.2) then
      if muestras_Turb[k]<0.1 then
      begin
        muestras_PE_0a10.pon_e(idx_0a10,muestras_PE[k]);
        muestras_Turb_0a10.pon_e(idx_0a10,muestras_Turb[k]);
        muestras_Pv3_0a10.pon_e(idx_0a10,muestras_pv3[k]);
        Inc(idx_0a10);
      end
      else
      begin
        muestras_PE_10a20.pon_e(idx_10a20,muestras_PE[k]);
        muestras_Turb_10a20.pon_e(idx_10a20,muestras_Turb[k]);
        muestras_Pv3_10a20.pon_e(idx_10a20,muestras_pv3[k]);
        Inc(idx_10a20);
      end
    else
    begin
      muestras_PE_Resto.pon_e(idx_resto,muestras_PE[k]);
      muestras_Turb_resto.pon_e(idx_resto,muestras_Turb[k]);
      muestras_Pv3_Resto.pon_e(idx_resto,muestras_pv3[k]);
      Inc(idx_resto);
    end;


  muestras_PE_0a10.resize(idx_0a10-1);
  muestras_Turb_0a10.resize(idx_0a10-1);
  muestras_Pv3_0a10.resize(idx_0a10-1);

  muestras_PE_10a20.resize(idx_10a20-1);
  muestras_Turb_10a20.resize(idx_10a20-1);
  muestras_Pv3_10a20.resize(idx_10a20-1);

  muestras_PE_Resto.resize(idx_resto-1);
  muestras_Turb_resto.resize(idx_resto-1);
  muestras_Pv3_Resto.resize(idx_resto-1);


  // muestras con intensidad de turbulencia desde 0 a 10 %
  sx := TSerieGrafico.Create('Pv3_0a10_ITurb', muestras_Pv3_0a10.toTDAOfNReal, nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('PE_muestras(0%-10% ITurb)', muestras_PE_0a10.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clGreen;
  sy.RadioMarcador := 2;
  sy.AnchoLineaMarcador := 1;
  g.AddSerie(sy);

  // muestras con intensidad de turbulencia desde 10 a 50 %
  sx := TSerieGrafico.Create('Pv3_10a50_ITurb', muestras_Pv3_10a20.toTDAOfNReal, nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('PE_muestras(10%-20% ITurb)', muestras_PE_10a20.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 2;
  sy.AnchoLineaMarcador := 1;
  g.AddSerie(sy);

  // muestras con intensidad de turbulencia desde inferior a 10 % o superior a 50 %
  sx := TSerieGrafico.Create('Pv3_Resto_ITurb', muestras_Pv3_Resto.toTDAOfNReal, nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('PE_muestras(Resto ITurb)', muestras_PE_Resto.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clblue;
  sy.RadioMarcador := 2;
  sy.AnchoLineaMarcador := 1;
  g.AddSerie(sy);


  NTramos := v_Pesos.n;

  //JFP: Grafico interavalos de potencias del viento
  for k := 1 to pv3_Representados.n do
  begin
    x:=pv3_Representados.e(k);
    linea:=TCajaGraficoXY.Create(x,0,x+0.5e-5,g.eje_y.v_max_forzado,'');
    g.AddCajaXY(linea);
  end;

  //Grafico Potencias electricas en pv3 de los representanes según las curvas "S".
  SetLength(pe_s_Rep,Representantes_pv3_medio.n);
  for k := 1 to Representantes_pv3_medio.n do
      pe_s_Rep[(k - 1)] := Calc_PE_curva_s(ang, Representantes_pv3_medio.e(k), ro_BASE);

  sx := TSerieGrafico.Create('Pv3Medio', Representantes_pv3_medio.toTDAOfNReal,
    nil, nil, '##.##', True);
  g.AddSerie(sx);

  sy := TSerieGrafico.Create('F0_s', pe_s_Rep, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 6;
  sy.AnchoLineaMarcador := 2;
  sy.eje := 0;
  g.AddSerie(sy);


  // Representantes - lineales
  setlength(v, 2 * NTramos);
  v[0] := pv3_Representados.e(1);
  for k := 1 to NTramos - 1 do
  begin
    v[2 * k - 1] := pv3_Representados.e(k + 1);
    v[2 * k] := pv3_Representados.e(k + 1);
  end;
  v[high(v)] := pv3_Representados.e(pv3_Representados.n);
  sx := TSerieGrafico.Create('pv3_representantes', v, nil, nil, '##.##', True);
  {$IFNDEF SOLO_GRAFICA_CURVAS_S}
  g.AddSerie(sx);
  {$ENDIF}
  setlength(v, 2 * NTramos);
  for k := 1 to NTramos do
  begin
    v[2 * (k - 1)] := v_a.e(k) + v_b.e(k) * v[2 * (k - 1)];
    v[2 * (k - 1) + 1] := v_a.e(k) + v_b.e(k) * v[2 * (k - 1) + 1];
  end;
  {$IFNDEF SOLO_GRAFICA_CURVAS_S}
  sy := TSerieGrafico.Create('Tramos', v, sx, g);
  sy.colorLinea := clNavy;
  sy.colorLineaMarcador := clNavy;
  sy.RadioMarcador := 4;
  sy.AnchoLineaMarcador := 2;
  g.AddSerie(sy);
  {$ENDIF}


  // ahora un plot de la interpolación de Hermit 10 puntos por tramo
  setlength(v, NTramos * 10);
  setlength(pe, length(v));
  setlength(pe_s, length(v));
  for k := 1 to NTramos do
  begin
    dx := pv3_Representados.e(k + 1) - pv3_Representados.e(k);
    for j := 1 to 10 do
    begin
      u := (j - 1) / 10.0;
      pv3 := pv3_Representados.e(k) + u * dx;
      v[(k - 1) * 10 + j - 1] := pv3;
      pe[(k - 1) * 10 + j - 1] := Calc_PE(ang, pv3, ro_BASE, Peso);
      pe_s[(k - 1) * 10 + j - 1] := Calc_PE_curva_s(ang, pv3, ro_BASE);
    end;
  end;

  sx := TSerieGrafico.Create('Pv3_Hermit', v, nil, nil, '##.##', True);
  g.AddSerie(sx);
  {$IFNDEF SOLO_GRAFICA_CURVAS_S}
  sy := TSerieGrafico.Create('PE_Hermit', pe, sx, g);
  sy.colorLinea := clAqua;
  sy.colorLineaMarcador := clAqua;
  sy.RadioMarcador := 1;
  sy.AnchoLineaMarcador := 1;
  sy.eje := 0;
  g.AddSerie(sy);
  {$ENDIF}

  //PS y JFP: Se grafican curvas S para cada sector.
  sy := TSerieGrafico.Create('PE_Curva_s', pe_s, sx, g);
  sy.colorLinea := clRed;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 0;
  sy.AnchoLineaMarcador := 1;
  sy.eje := 0;
  g.AddSerie(sy);



  // muestra
  sx := TSerieGrafico.Create('Pv3Medio', Representantes_pv3_medio.toTDAOfNReal,
    nil, nil, '##.##', True);
  g.AddSerie(sx);
  sy := TSerieGrafico.Create('Peso', v_Pesos.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clFuchsia;
  sy.RadioMarcador := 3;
  sy.AnchoLineaMarcador := 1;
  sy.eje := 1;
  g.AddSerie(sy);

  {$IFNDEF SOLO_GRAFICA_CURVAS_S}
  sy := TSerieGrafico.Create('F0', v_F0.toTDAOfNReal, sx, g);
  sy.colorLinea := clNone;
  sy.colorLineaMarcador := clRed;
  sy.RadioMarcador := 3;
  sy.AnchoLineaMarcador := 2;
  sy.eje := 0;
  g.AddSerie(sy);
  {$ENDIF}

  g.SaveJPG(CarpetaResultados_modelado_RO( Parque_MT.nid_Central) +
    DIrectorySeparator + IntToStr(trunc(ang * 100)) + '_' + IntToStr(
    cnt_pasda)+'_'+ IntToStr(trunc(Alfa_velFilt*100)) + '.jpg', 2100, 1000);
  g.Free;

  muestras_PE_Resto.Free;
  muestras_Pv3_Resto.Free;
  muestras_Turb_resto.Free;

  muestras_PE_0a10.Free;
  muestras_Pv3_0a10.Free;
  muestras_Turb_0a10.Free;

  muestras_PE_10a20.Free;
  muestras_Pv3_10a20.Free;
  muestras_Turb_10a20.Free;

end;

{$ENDIF}


procedure guarde_problema(x, y: TVectR; Ymax, curva_s_alfa, curva_s_vm: NReal);
var
  f: textfile;
  k: integer;

begin
  assignfile(f, 'c:\basura\curva_s.xlt');
  rewrite(f);
  writeln(f, 'x', #9, 'y', #9, 's');
  for k := 1 to x.n do
  begin
    writeln(f, x.e(k), #9, YMax / (1 + exp(y.e(k))), #9, YMax /
      (1 + exp(-curva_s_alfa * (x.e(k) - curva_s_vm))));
  end;
  closefile(f);

end;

procedure TParque_ModeloEmpirico.Calc_Curva_s_cluser_ang(
  var curva_s_alfa, curva_s_vm, curva_s_errc2: NReal; var curva_s_cntMuestras: integer;
  pv3Anemo_ClusterAng, pe_MWClusterAng: TVectR; Parque_MT_Curva_s: TCurvaAeroGen);

var
  x, y: TVectR;
  k: integer;
  PE: NReal;
  cnt: integer;
  Pmax, P01, P90,P80: NReal;
  a, b: NReal;
  v: NReal;
begin
  x := TVectR.Create_init(pv3Anemo_ClusterAng.n);
  y := TVectR.Create_init(pe_MWClusterAng.n);
  cnt := 0;
  PMax := Parque_MT_Curva_s.PMax;
  P01 := PMax * 0.05;
  //P90 := PMax * 0.95;
  P80 := PMax * 0.8;
  for k := 1 to x.n do
  begin
    PE := pe_MWClusterAng.e(k);
    if (PE >= P01) and (PE <= P80) then
    begin
      v := vOfpv3(pv3Anemo_ClusterAng.e(k), ro_BASE);
      Inc(cnt);
      x.pon_e(cnt, v);
      y.pon_e(cnt, ln(PMax / PE - 1));
    end;
  end;
  x.resize(cnt);
  y.resize(cnt);
  curva_s_cntMuestras := cnt;
  if cnt > 10 then
  begin
    try
      curva_s_errc2 := y.AproximacionLineal(a, b, x);
      curva_s_alfa := -a;
      curva_s_vm := b / curva_s_alfa;
      guarde_problema(x, y, PMax, curva_s_alfa, curva_s_vm);
    except
      on E: Exception do
      begin
        curva_s_alfa := -1;
        curva_s_vm := -1;
      end;
    end;
  end
  else
  begin
    curva_s_alfa := -1;
    curva_s_vm := -1;
  end;
  x.Free;
  y.Free;
end;

// Ordenamos las puestras por pv3 y luego para cada valor de pv3 desde 0
// al valor correspondiete  a 20 mps buscamos los representantes
// sin no hay representantes válidos pondremos -1 en el peso
// para indicar que ese punto no debe usarse.
procedure TParque_ModeloEmpirico.Representar_Clusters(NSectores: integer;
  // cantidad de sectores en que se clusteriza
  NTramosRepresentados: integer; // Tramos de v en que se realiza la representación
  overlapping: NReal; // overlapping lateral de los sectores. 0.2 = 20% para cada lado
  cnt_Pasada: integer);
var
  kRep: integer;
  ang: NReal;
  dang: NReal;
  dvel: NReal;
  vel: NReal;
  kAng: integer;
  idx: TDAOfNInt;
  v_Pesos, v_a, v_b, v_F0: TVectR;
  pv3a, pv3b: NReal;
  ja, jb: integer;
  j: integer;
  peso: integer;
  pv3, peMW: NReal;
  VelMax, epsilon: NReal;
  mcA, mcB: TMatR;
  mcA_Invertible: boolean;
  exp10: integer;
  coef_a, coef_b: NReal;
  Matriz_alfa_Borde, Matriz_alfa_completa, Matriz_vm_Borde, Matriz_vm_completa: TMatR;
  cnt_iter, iter: integer;
  Max_cnt_iter: integer;
  i: integer;
  f: textfile;
  pv3Anemo_ClusterAng_Mat, pe_MW_ClusterAng_Mat,I_Turb_ClusterAng_Mat: TDAOfDAofNReal;

begin

  if cnt_Pasada > 1 then
  begin
    pv3_Representados.Free;
    Representantes_pv3_medio.Free;
    peso_Representantes.Free;
    pe_MW_Representantes_a.Free;
    pe_MW_Representantes_b.Free;
    Representante_FO.Free;
    curva_s_alfa.Free;
    curva_s_vm.Free;
    curva_s_errc2.Free;
    curva_s_cntMuestras.Free;
  end;

  // Primero punto = v = 0; Ultimo Punto = v_max
  pv3_Representados := TVectR.Create_Init(NTramosRepresentados + 1);
  // dang = medio sector
  dang := 360 / NSectores / 2 * (1 + overlapping);

  //  pv3_v20mps:= pv3Ofv( 20, 1.2 );

  // Distribución inicial de representantes
  VelMax := sVel_mps.maxVal;
  dvel := VelMax / NTramosRepresentados;
  for kRep := 1 to NTramosRepresentados + 1 do
  begin
    vel := (kRep - 1) * dvel;
    pv3_Representados.pon_e(kRep, pv3Ofv(vel, ro_BASE));
  end;

  Representantes_pv3_medio := TVectR.Create_Init(NTramosRepresentados);
  for kRep := 1 to NTramosRepresentados do
    Representantes_pv3_medio.pon_e(
      krep, (pv3_Representados.e(krep) + pv3_Representados.e(krep + 1)) / 2.0);

  peso_Representantes := TMatR.Create_Init(NSectores, NTramosRepresentados);
  pe_MW_Representantes_a := TMatR.Create_Init(NSectores, NTramosRepresentados);
  pe_MW_Representantes_b := TMatR.Create_Init(NSectores, NTramosRepresentados);
  Representante_FO := TMatR.Create_Init(NSectores, NTramosRepresentados);

  SetLength(pv3Anemo_ClusterAng_Mat, NSectores + 1);
  SetLength(pe_MW_ClusterAng_Mat, NSectores + 1);
  SetLength(I_Turb_ClusterAng_Mat, NSectores + 1);

  curva_s_alfa := TVectR.Create_init(NSectores);
  curva_s_vm := TVectR.Create_init(NSectores);
  curva_s_errc2 := TVectR.Create_init(NSectores);
  curva_s_cntMuestras := TVectE.Create_init(NSectores);

  /////////////////////////////////////////////////////////////////////////////
  //Interpolación de los alfa y vm que están bajo el umbral de muestras.
  //El umbral de muestras es de 50 muestras por cluster, si el cluster tiene
  //menos de 50 muestras, se considera que el alfa y vm deben ser calculados
  //mediante interpolación con sus vecinos
  ////////////////////////////////////////////////////////////////////////////

  Matriz_alfa_Borde := TMatR.Create_Init(NSectores, 1);
  Matriz_alfa_Borde.Ceros;

  Matriz_alfa_completa := TMatR.Create_Init(NSectores, 1);
  Matriz_alfa_completa.Ceros;

  Matriz_vm_Borde := TMatR.Create_Init(NSectores, 1);
  Matriz_vm_Borde.Ceros;

  Matriz_vm_completa := TMatR.Create_Init(NSectores, 1);
  Matriz_vm_completa.Ceros;
  /////////////////////////////////////////////////////////////




  for kAng := 1 to NSectores do
  begin
    ang := 360 / NSectores * (kAng - 1);  // ángulo central del sector

    // Carga en pv3ANemo_ClusterAng,pe_MW_ClusterAng,I_Turb_ClusterAng
    //las muestras del sector [ang-dang, ang+dang]
    Cluster_ang(ang, dang);

    // reeordenamos las muestras por pv3Anemo creciente
    setlength(idx, pv3Anemo_ClusterAng.n + 1);
    for j := 0 to high(idx) do
      idx[j] := j;
    pv3Anemo_ClusterAng.Sort_idx(True, idx);
    pe_MW_ClusterAng.SortByIdx(idx);
    I_Turb_ClusterAng.SortByIdx(idx);

    pv3Anemo_ClusterAng_Mat[kAng] := pv3Anemo_ClusterAng.toTDAOfNReal;
    pe_MW_ClusterAng_Mat[kAng] := pe_MW_ClusterAng.toTDAOfNReal;
    I_Turb_ClusterAng_Mat[kAng]:=I_Turb_ClusterAng.toTDAOfNReal;

    Calc_Curva_s_cluser_ang(
      curva_s_alfa.pv[kAng],
      curva_s_vm.pv[kAng],
      curva_s_errc2.pv[kAng],
      curva_s_cntMuestras.pv[kAng],
      pv3Anemo_ClusterAng, pe_MW_ClusterAng, Parque_MT.Curva_s);

    // indexamos el vector de pesos y de coeficientes de la aprox. lineal del tramo
    v_Pesos := peso_Representantes.Fila(kAng);
    v_a := pe_MW_Representantes_a.Fila(kAng);
    v_b := pe_MW_Representantes_b.Fila(kAng);
    v_F0 := Representante_FO.Fila(kAng);


    ////////////////////////////////////////
    if curva_s_cntMuestras.pv[kAng] > 50 then
    begin
      Matriz_alfa_Borde.pon_e(kAng, 1, curva_s_alfa.e(kAng));
      Matriz_vm_Borde.pon_e(kAng, 1, curva_s_vm.e(kAng));
    end;
    ///////////////////////////////////////

    for kRep := 1 to NTramosRepresentados do
    begin
      // extremos del intervalo
      pv3a := pv3_Representados.e(kRep);
      pv3b := pv3_Representados.e(kRep + 1);

      ja := Ceil(pv3Anemo_ClusterAng.inv_interpol(pv3a));
      jb := Floor(pv3Anemo_ClusterAng.inv_interpol(pv3b));

      ja := min(max(1, ja), pv3Anemo_ClusterAng.n);
      jb := min(max(1, jb), pv3Anemo_ClusterAng.n);

      if jb > ja then
      begin
        peso := (jb - ja + 1);

        // mínimos cuadrados clásico
        mcA := TMatR.Create_Init(2, 2);
        mcA.Ceros;
        mcB := TMatR.Create_Init(2, 1);
        mcB.Ceros;
        for j := ja to jb do
        begin
          peMW := pe_MW_ClusterAng.e(j);
          pv3 := pv3Anemo_ClusterAng.e(j);
          mcA.acum_e(1, 2, pv3);
          mcA.acum_e(2, 2, pv3 * pv3);
          mcB.acum_e(1, 1, peMW);
          mcB.acum_e(2, 1, pv3 * peMW);
        end;
        mcA.pon_e(1, 1, peso);
        mcA.pon_e(2, 1, mcA.e(1, 2));
        mcA.Escaler(mcB, mcA_Invertible, exp10);
        if not mcA_Invertible then
          raise Exception.Create(
            'Atención!!!! sistema lineal no invertible en la aproximación de un tramo!!!');
        coef_a := mcB.e(1, 1);
        coef_b := mcB.e(2, 1);
        v_Pesos.pon_e(kRep, peso);
        v_a.pon_e(kRep, coef_a);
        v_b.pon_e(kRep, coef_b);
        mcA.Free;
        mcB.Free;
      end
      else
      begin
        v_Pesos.pon_e(kRep, -1);
        v_a.pon_e(kRep, 0);
        v_b.pon_e(kRep, 0);
      end;
      v_F0.pon_e(kRep, v_a.e(kRep) + v_b.e(kRep) * Representantes_pv3_medio.e(kRep));
    end;

  end;


  /////////////////////////////////////////////
  Rellenar_matriz(Matriz_alfa_completa, Matriz_alfa_Borde, epsilon, iter);
  Rellenar_matriz(Matriz_vm_completa, Matriz_vm_Borde, epsilon, iter);

  for kAng := 1 to NSectores do
  if curva_s_cntMuestras.pv[kAng] <= 50 then
  begin
    curva_s_alfa.pon_e( kAng, Matriz_alfa_completa.e(kAng, 1 ));
    curva_s_vm.pon_e( kAng, Matriz_vm_completa.e(kAng, 1 ));
  end;
  ///////////////////////////////////////


  {$IFDEF CON_GRAFICOS}


  for kAng := 1 to NSectores do
    if pv3Anemo_ClusterAng.n > 3 then
      PLOT_GRAFICAS_DEL_CLUSTER_ANG(kAng, pv3_Representados, Representantes_pv3_medio,
        pv3Anemo_ClusterAng_Mat[kAng], pe_MW_ClusterAng_Mat[kAng],I_Turb_ClusterAng_Mat[kAng],
        peso_Representantes.Fila(kAng), pe_MW_Representantes_a.Fila(kAng),
        pe_MW_Representantes_b.Fila(kAng), Representante_FO.Fila(kAng)
        , Matriz_alfa_completa.e(kAng, 1), Matriz_vm_completa.e(kAng, 1), cnt_Pasada);

  {$ENDIF}

  //writeln('epsilon',epsilon,'  iter',iter);
  //MatrizFactor.WriteArchiXLT('c:\basura\MatrizFactor.xlt');

  //escribo resultados//
  assignfile(f, CarpetaResultados_modelado_RO(parque_MT.nid_Central) +
    '\curvas_s_completadas_.xlt');
  rewrite(f);
  writeln(f, 'ang', #9, 'alfa_completo', #9, 'vm_completo');
  //, #9, 'errc2', #9, 'cntMuestras', #9, 'MT_alfa', #9, 'MT_vm' );

  for kAng := 1 to NSectores do
  begin
    ang := AngOfjSector(kAng);
    writeln(f, ang, #9, Matriz_alfa_completa.e(kAng, 1), #9, Matriz_vm_completa.e(kAng, 1));
  end;
  closefile(f);
  ///////////////////////////////////////////////////////////////
end;




function TParque_ModeloEmpirico.Calc_PE(ang, pv3Anemo, ro: NReal;
  var Peso: NReal): NReal;
var
  kang1, kang2: integer;
  rAng: NReal;
  a1, a2: NReal;
  jpv3: NReal;
  PE1, PE2: NReal;
  peso1, peso2: NReal;

  u: NREal;
  kTramo: integer;
  NTramos: integer;
  dx: NReal;

begin

  SectorOfAng(kAng1, kAng2, a1, a2, Ang);


  jpv3 := Representantes_pv3_medio.inv_interpol(pv3Anemo);
  kTramo := trunc(jpv3);
  if kTramo < 1 then
  begin
    // Extremo inferior
    Peso1 := peso_Representantes.e(kAng1, 1);
    Peso2 := peso_Representantes.e(kAng2, 1);
    PE1 := Representante_FO.e(kAng1, 1) + pe_MW_Representantes_b.e(
      kang1, 1) * (pv3Anemo - Representantes_pv3_medio.e(1));
    PE2 := Representante_FO.e(kAng2, 1) + pe_MW_Representantes_b.e(
      kang2, 1) * (pv3Anemo - Representantes_pv3_medio.e(1));
  end
  else if kTramo >= Representantes_pv3_medio.n then
  begin
    // Extremo superior
    NTramos := peso_Representantes.nc;
    Peso1 := peso_Representantes.e(kAng1, NTramos);
    Peso2 := peso_Representantes.e(kAng2, NTramos);
    PE1 := Representante_FO.e(kAng1, NTramos) + pe_MW_Representantes_b.e(
      kang1, NTramos) * (pv3Anemo - Representantes_pv3_medio.e(NTramos));
    PE2 := Representante_FO.e(kAng2, NTramos) + pe_MW_Representantes_b.e(
      kang2, NTramos) * (pv3Anemo - Representantes_pv3_medio.e(NTramos));
  end
  else
  begin
    // Interpol Hermit común
    u := frac(jpv3);
    dx := Representantes_pv3_medio.e(kTramo + 1) - Representantes_pv3_medio.e(kTramo);

    PE1 := Calc_PE_hermit(u, dx, peso1, kTramo, peso_Representantes.Fila(
      kAng1), Representante_FO.Fila(kAng1), pe_MW_Representantes_b.Fila(kang1));

    PE2 := Calc_PE_hermit(u, dx, peso2, kTramo, peso_Representantes.Fila(
      kAng2), Representante_FO.Fila(kAng2), pe_MW_Representantes_b.Fila(kang2));
  end;

  if ((a1 > 0.001) and (peso1 < UmbralPoblacionMinimaCluster)) or
    ((a2 > 0.001) and (peso2 < UmbralPoblacionMinimaCluster)) then
  begin
    peso := -1;
    Result := 0;
  end
  else
  begin
    Peso := a1 * Peso1 + a2 * Peso2;
    Result := (a1 * PE1 + a2 * PE2) * Ajuste_Cp_ro(ro);
  end;
end;

////////////////////////
function TParque_ModeloEmpirico.Calc_PE_curva_s(ang, pv3Anemo, ro: NReal): NReal;
var
  kang1, kang2: integer;
  a1, a2, v3: NReal;
  PE1, PE2: NReal;
begin
  SectorOfAng(kAng1, kAng2, a1, a2, Ang);
  v3 := vOfpv3(pv3Anemo, ro);
  PE1 := 1 / (1 + exp(-self.curva_s_alfa.e(kAng1) *
    (v3 - self.curva_s_vm.e(kAng1))));
  PE2 := 1 / (1 + exp(-self.curva_s_alfa.e(kAng2) *
    (v3 - self.curva_s_vm.e(kAng2))));
  Result := Parque_MT.Curva_s.PMax*(a1 * PE1 + a2 * PE2) * Ajuste_Cp_ro(ro);
end;
/////////////////////////



(***
Fórmula según resumen de Cuaderno_rojo rch-21/10/2004

y0 : valor de la función al inicio del intervalo
dydx0: valor de la derivada al inicio del intervalo
y1: valor de la función al final del intervalo
dydx1: valor de la deriva al final del intervalo
dx: ancho del intervalo
u: [0, 1] valor donde se quiere calcular y(u) expresado en p.u. de dx.

**)
function hermit(const y0, dydx0, y1, dydx1, dx, u: double): NReal;
var
  a, b, ac, bc, h: NReal;
begin
  a := u * u * (2 * u - 3) + 1;
  ac := 1 - a;
  h := (1 - u) * u;
  b := h * (1 - u);
  bc := -h * u;
  Result := a * y0 + ac * y1 + (b * dydx0 + bc * dydx1) * dx;
end;


function TParque_ModeloEmpirico.Calc_PE_hermit(u, dx: NReal; var Peso: NReal;
  kTramo: integer; rep_Peso, rep_F0, rep_b: TVectR): NReal;

var
  PE: NReal;
  peso1, peso2: NReal;
  F0, F1, d0, d1: NReal;
begin

  peso1 := rep_Peso.e(ktramo);
  peso2 := rep_Peso.e(ktramo + 1);
  if ((u < 0.999) and (peso1 < UmbralPoblacionMinimaCluster)) or
    ((u > 0.001) and (peso2 < UmbralPoblacionMinimaCluster)) then
  begin
    peso := -1;
    Result := 0;
    exit;
  end;

  F0 := Rep_F0.e(ktramo);
  d0 := Rep_b.e(ktramo);
  F1 := Rep_F0.e(ktramo + 1);
  d1 := Rep_b.e(ktramo + 1);

  PE := hermit(F0, d0, F1, d1, dx, u);
  peso := peso1 * (1 - u) + peso2 * u;
  Result := PE;
end;



function angEnCluster(ang, ang0, dang: NReal): boolean;
var
  d: NReal;
begin
  d := abs(ang - ang0);
  while d > 360 do
    d := d - 360;
  Result := (d <= dang);
end;

// Filtra los datos en fijando en pv3Anemo_ClusterAng y pe_MW_ClusterAng
// la potencia del viento en el Anemo y la potencia Eléctrica correspondiente
// a las muestas que se encuentran  abs(ang - ang0) < dang
procedure TParque_ModeloEmpirico.Cluster_Ang(ang0, dang: NReal);
var
  kDato: integer;
  jDato: integer;
  ang, ro, vel, potE: NReal;
  pv3: NReal;
begin
  if pv3Anemo_ClusterAng <> nil then
    pv3Anemo_ClusterAng.Free;
  pv3Anemo_ClusterAng := TVectR.Create_Init(NDatos);

  if pe_MW_ClusterAng <> nil then
    pe_MW_ClusterAng.Free;
  pe_MW_ClusterAng := TVectR.Create_Init(NDatos);

  if I_Turb_ClusterAng <> nil then
    I_Turb_ClusterAng.Free;
  I_Turb_ClusterAng := TVectR.Create_Init(NDatos);

  jDato := 0;
  for kDato := 1 to NDatos do
  begin
    if filtro_huecos_vdp[kDato] or filtro_coherencia[kDato] or
      filtro_cgm[kDato] or filtro_distancia[kDato] then
      continue;

    ang := self.sDir_gr.e(kDato);
    if angEnCluster(ang, ang0, dang) then
    begin
      Inc(jdato);

      vel := sVel_mps.e(kDato);
      ro := ro_aire(sTemp_grC.e(kDato), sPre_hPa.e(kDato), sHumRel_p100.e(kDato));

      pv3 := pv3Ofv(vel, ro);

      potE := sPot_MW.e(kDato) / Ajuste_Cp_ro(ro);

      pv3Anemo_CLusterAng.pon_e(jDato, pv3);
      pe_MW_ClusterAng.pon_e(jDato, potE);
      //JFP: agrego turbulencia a las muestras del clúster
      I_Turb_ClusterAng.pon_e(jDato, sI_Turb.e(kDato));
    end;
  end;
  pv3Anemo_CLusterAng.Resize(jDato);
  pe_MW_ClusterAng.Resize(jDato);
  I_Turb_ClusterAng.resize(jDato);
end;



end.
