unit uReferenciaMonArchivo;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  uReferenciaMonitor, uEventosOptSim,
  uVarDefs, uMonitorArchivo, uCosa, uCosaConNombre,
  Classes, SysUtils, uMonitores,
  uSalasDeJuego;

resourcestring
  rsMonitorDeArchivo = 'Monitor de Archivo';
  mesEnElMonitor = 'En el monitor ';

type

  { TReferenciaMonArchivo }

  TReferenciaMonArchivo = class(TReferenciaMonMultiVar)
  public

  (**************************************************************************)
  (*              A T R I B U T O S   P E R S I S T E N T E S               *)
  (**************************************************************************)

     filename: string;
     {$IFDEF MIGRACION_PERSISTENCIA}
       mensajes: TEventoOptSimMsgs;
     {$ENDIF}
  (**************************************************************************)

  {$IFNDEF MIGRACION_PERSISTENCIA}
     mensajes: TDAofMsgs;
  {$ENDIF}


    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create(capa: integer; nombreMon: string; filename: string;
      defVars: TListaDeCosas; xEventosProcs: TDAOfEventoProc; msgs: TDAofMsgs);
    {$ENDIF}
    {$IFDEF MIGRACION_PERSISTENCIA}
    constructor Create(capa: integer; nombreMon: string; filename: string;
      defVars: TListaDeCosas; xEventosProcs: TDAOfEventoProc; msgs: TEventoOptSimMsgs);
    {$ENDIF}

    {$IFNDEF MIGRACION_PERSISTENCIA}
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    {$ENDIF}

    class function DescClase: string; override;
    procedure crearMonitorDeVariables(cosasMonitoreables: TListaDeCosasConNombre);
      override;
    class function tiposMonitoreables: AOfVarDef; override;

    {$IFDEF MIGRACION_PERSISTENCIA}
    class function CreateDataColumnList(xClaseDeCosa:TClaseDeCosa; xVersion: Integer = -2): TDataColumnListOfCosa; virtual;
  published
    property _filename: string read filename write filename;
    property _mensajes: TEventoOptSimMsgs read mensajes write mensajes;
    {$ENDIF}
  end;

procedure AlInicio;
procedure AlFinal;

implementation

var
  VarDefsValidas_TMonArchivo: AOfVarDef;

//--------------------------------
//Metodos de TReferenciaMonArchivo
//================================

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TReferenciaMonArchivo.Create( capa: integer; nombreMon: string; filename: string;
  defVars: TListaDeCosas {of TReferenciaDefVar}; xEventosProcs: TDAOfEventoProc;
  msgs: TDAofMsgs);
{$ENDIF}
{$IFDEF MIGRACION_PERSISTENCIA}
constructor TReferenciaMonArchivo.Create( capa: integer; nombreMon: string; filename: string;
  defVars: TListaDeCosas {of TReferenciaDefVar}; xEventosProcs: TDAOfEventoProc;
  msgs: TEventoOptSimMsgs);
{$ENDIF}
var
  j: TEventoOptSim;
begin
  inherited Create(capa, nombreMon, xEventosProcs, defVars);
  {$IFNDEF MIGRACION_PERSISTENCIA}
    for j := low(TEventoOptSim) to high(TEventoOptSim) do
      mensajes[j] := msgs[j];
  {$ENDIF}
  {$IFDEF MIGRACION_PERSISTENCIA}
    mensajes:= msgs;
  {$ENDIF}
  Self.filename := filename;
end;

{$IFNDEF MIGRACION_PERSISTENCIA}
constructor TReferenciaMonArchivo.Create_ReadFromText(f: TArchiTexto);
var
  cadena, msg: string;
  evento: TEventoOptSim;
  i: TEventoOptSim;
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('filename', filename);
  f.EjecutarLectura;
  for i := low(mensajes) to high(mensajes) do
  begin
    f.IniciarLecturaRetrasada;
    f.rd('evento', cadena);
    f.rd('mensaje', msg);
    f.EjecutarLectura;
    evento := StrToEvento(cadena);
    mensajes[evento] := msg;
  end;
end;
{$ENDIF}

{$IFNDEF MIGRACION_PERSISTENCIA}
procedure TReferenciaMonArchivo.WriteToText(f: TArchiTexto);
var
  j: TEventoOptSim;
begin
  inherited WriteToText(f);
  f.wr('filename', filename);

  for j := low(mensajes) to high(mensajes) do
  begin
    f.wr('evento', EventoToStr(j));
    f.wr('mensaje', mensajes[j]);
  end;
end;
{$ENDIF}

class function TReferenciaMonArchivo.DescClase: string;
begin
  Result := rsMonitorDeArchivo;
end;

procedure TReferenciaMonArchivo.crearMonitorDeVariables(cosasMonitoreables:
  TListaDeCosasConNombre);
var
  ListaDefVars: TList; {of TVarDef}
  i: integer;
  varDef: TVarDef;
  errores: string;
begin
  eventos := sortTDAOfEventoProcByProc(eventos);
  errores := '';
  ListaDefVars := TList.Create;
  ListaDefVars.Capacity := defVars.Count;
  for i := 0 to defVars.Count - 1 do
  begin
    try
      begin
        varDef := TReferenciaDefVar(defVars[i]).resolver(cosasMonitoreables);
        varDef.Prepararse;
        ListaDefVars.Add(varDef);
      end
    except
      on E: EReferenciaDefVarException do
      begin
        errores := errores + E.Message;
      end
    end;
  end;
  i := ListaDefVars.Count;
  if ListaDefVars.Count <> 0 then
    monitor := TMonArchivo.Create(ListaDefVars, filename, mensajes);
  ListaDefVars.Free;
  if errores <> '' then
    raise EMonitorException.Create(errores + mesEnElMonitor + nombre + '.', i);
end;

class function TReferenciaMonArchivo.tiposMonitoreables: AOfVarDef;
begin
  Result := VarDefsValidas_TMonArchivo;
end;

{$IFDEF MIGRACION_PERSISTENCIA}
class function TReferenciaMonArchivo.CreateDataColumnList(
  xClaseDeCosa: TClaseDeCosa; xVersion: Integer): TDataColumnListOfCosa;
begin

  if (xVersion<>-2) then
    raise Exception.Create('Está intentando utilizar un archivo *.ese no soportado en la versión actual del sistema. \n\r' +
                           'MENSAJE INTERNO: \n\r' +
                           '---------------- \n\r' +
                           'TReferenciaMonArchivo.CreateDataColumnList: Es necesario cambiar la forma en que se representa el array de mensajes en el archivo *.ese');

  Result:=inherited CreateDataColumnList(xClaseDeCosa, xVersion);
  Result.AddStringColumn('filename','');
  Result.AddCosaColumn('mensajes');
end;
{$ENDIF}

procedure AlInicio;
begin
  registrarClaseDeCosa(TReferenciaMonArchivo.ClassName, TReferenciaMonArchivo);

  SetLength(VarDefsValidas_TMonArchivo, 1);
  VarDefsValidas_TMonArchivo[0] := TVarDef;
end;

procedure AlFinal;
begin
  SetLength(VarDefsValidas_TMonArchivo, 0);
end;

end.
 
